#include "I2Cdev.h"
#include "MPU6050_6Axis_MotionApps20.h"
#include "Wire.h"

/*
 * 姿态传感器MPU6050将自行进行姿态融合
 * 只需要将姿态读取即可
 */

//姿态感应传感器MPU6050安装误差补偿参数
//代表小车在平衡位置时的传感器读数
const float offset_angle = 0;

float angle;   //当前倾斜角度
int16_t gyro;  //当前陀螺仪值
MPU6050 mpu;

// MPU control/status vars
bool dmpReady = false;  // set true if DMP init was successful
uint8_t mpuIntStatus;   // holds actual interrupt status byte from MPU
uint8_t devStatus;    // return status after each device operation (0 = success,
                      // !0 = error)
uint16_t packetSize;  // expected DMP packet size (default is 42 bytes)
uint16_t fifoCount;   // count of all bytes currently in FIFO
uint8_t fifoBuffer[64];  // FIFO storage buffer

Quaternion q;

VectorFloat gravity;

uint8_t* getMPUBuff() {
    if (!dmpReady) {
        Serial.println("start init MPU");
        Wire.begin();
        Wire.setClock(100000);
        mpu.initialize();
        devStatus = mpu.dmpInitialize();
        if (devStatus == 0) {
            mpu.setDMPEnabled(true);
            mpuIntStatus = mpu.getIntStatus();
            dmpReady = true;
            packetSize = mpu.dmpGetFIFOPacketSize();
        }
        return NULL;
    }

    mpuIntStatus = mpu.getIntStatus();
    fifoCount = mpu.getFIFOCount();
    if ((mpuIntStatus & 0x10) || fifoCount > 512) {
        mpu.resetFIFO();
        Serial.println("FIFO Overflow");
    } else if (mpuIntStatus & 0x02) {
        while (fifoCount < packetSize) fifoCount = mpu.getFIFOCount();
        mpu.getFIFOBytes(fifoBuffer, packetSize);
        mpu.resetFIFO();
        return fifoBuffer;
    }
    return NULL;
}

bool updateIMU() {
    uint8_t* buff = getMPUBuff();
    if (buff == NULL) return false;
    mpu.dmpGetQuaternion(&q, buff);
    mpu.dmpGetGravity(&gravity, &q);
    //正面安装
    // angle = atan2(gravity.y, gravity.z) * -180 / M_PI - offset_angle;
    // gyro = (buff[16] << 8) | buff[17];  // x轴陀螺仪读数
    //反面安装
    angle = atan2(gravity.y, -gravity.z) * 180 / M_PI - offset_angle;
    int16_t t[3];
    gyro = -((buff[16] << 8) | buff[17]);  // x轴陀螺仪读数
    return true;
}

/*
 * 反面安装时,姿态传感器需要一段时间才能找到正确的角度
 */