/*
 * 遥控控制
 * 更新遥控器信息,根据遥控信息,设置偏置变量,从而达到控制效果
 */

#include <Arduino.h>

int rc_forward = 0;  //前进/后退期望
int rc_truen = 0;    //转弯期望

bool enableGravity = false;     //是否启用重力控制
bool enableUltrasonic = false;  //是否启用超声波避障模式

char newButtonState;
char oldButtonState;
unsigned long incomeTimeout;
int left_X = 0;
int left_Y = 0;
int right_X = 0;
int right_Y = 0;
int gravity_X = 0;
int gravity_Y = 0;
int gravity_Z = 0;

#define button_L (1)
#define button_R (2)
#define button_U (4)
#define button_D (8)
#define button_A (16)
#define button_B (32)
#define button_X (64)
#define button_Y (128)

void remoteUpdate() {
    if (millis() - incomeTimeout > 500) {
        //遥控器500ms内没有信号时,保持直立
        rc_forward = rc_truen = 0;
    } else {
        char buttonChange = newButtonState ^ oldButtonState;
        char buttonDown = buttonChange & newButtonState;
        char buttonUp = buttonChange & (~newButtonState);
        if (buttonUp & button_X) {
            enableGravity = !enableGravity;
            if (enableGravity) {
                Serial.println("重力感应控制已开启");
            } else {
                Serial.println("重力感应控制已关闭");
            }
        }
        if (buttonUp & button_Y) {
            enableUltrasonic = !enableUltrasonic;
            if (enableUltrasonic) {
                Serial.println("避障模式已开启");
            } else {
                Serial.println("避障模式已关闭");
            }
        }

        if (newButtonState & button_D) {
            rc_forward = -200;
            rc_truen = 0;
        } else if (newButtonState & button_L) {
            rc_forward = 0;
            rc_truen = -150;
        } else if (newButtonState & button_R) {
            rc_forward = 0;
            rc_truen = 150;
        } else if (newButtonState && button_U) {
            rc_forward = 250;
            rc_truen = 0;
        } else {
            rc_truen = left_X;
            rc_forward = left_Y;

            //任何遥控信号都优先于避障模式/遥控器重力控制
            //因为遥控器重力通道总是有数据,无法分辨是否人为在控制
            //所以避障模式优先于重力控制
            if (rc_truen == 0 && rc_forward == 0) {
                if (enableUltrasonic) {
                    //超声波最大检查距离500mm
                    // Arduino中检查脉冲宽度受中断函数影响,而此程序的中断非常频繁
                    //所以超声波功能将不太可靠
                    //好在,此影响只会导致测量距离偏短,增加误触发机率.
                    if (ultrPing < 300) {
                        rc_truen = 150;  //模拟遥控器值,转弯
                        rc_forward = 0;
                    } else {
                        rc_forward = 150;  //模拟遥控器值,直行
                        rc_truen = 0;
                    }
                } else if (enableGravity) {
                    rc_forward = -gravity_Y;
                    rc_truen = gravity_X;
                }
            }
            rc_forward = constrain(rc_forward, -255, 255);
            rc_truen = constrain(rc_truen, -255, 255);
        }
    }
}
//此函数在进入loop循环前由Arduino调用(如果串口有数据)
void serialEvent() {
    static char letter = 0;
    static int number = 0;
    static char minusSign = 0;
    while (Serial.available()) {
        char data = Serial.read();
        incomeTimeout = millis();
        if ((data >= '0') && (data <= '9')) {
            number *= 10;
            number += data - '0';
        } else if (data == '-') {
            minusSign = 1;
        } else {
            if (minusSign) number = -number;
            switch (letter) {
                case 'u':
                    left_X = number;
                    break;
                case 'v':
                    left_Y = number;
                    break;
                case 'w':
                    right_X = number;
                    break;
                case 'a':
                    right_Y = number;
                    break;
                case 'x':
                    gravity_X = number;
                    break;
                case 'y':
                    gravity_Y = number;
                    break;
                case 'z':
                    gravity_Z = number;
                    break;
                default:
                    break;
            }
            number = 0;
            minusSign = 0;
            switch (data) {
                case 'L':
                    newButtonState |= button_L;
                    break;
                case 'R':
                    newButtonState |= button_R;
                    break;
                case 'U':
                    newButtonState |= button_U;
                    break;
                case 'D':
                    newButtonState |= button_D;
                    break;
                case 'A':
                    newButtonState |= button_A;
                    break;
                case 'B':
                    newButtonState |= button_B;
                    break;
                case 'X':
                    newButtonState |= button_X;
                    break;
                case 'Y':
                    newButtonState |= button_Y;
                    break;
                case '\n':
                    remoteUpdate();
                    oldButtonState = newButtonState;
                    newButtonState = 0;
                    break;
                case 'p':
                case 'P':
                    if (letter == '\n' || letter == 0 || letter == '\r') {
                        parseParameters();
                        letter = 0;
                    }
                default:
                    break;
            }
            letter = data;
        }
    }
}
