#include "Wire.h"

/*
 * 姿态传感器MPU6050将自行进行姿态融合
 * 只需要将姿态读取即可
 */

#include <Arduino.h>
#include "Wire.h"

//姿态感应传感器MPU6050安装误差补偿参数
//代表小车在平衡位置时的传感器读数
const float offset_angle = -3;


//卡尔曼滤波相关参数
float Q_angle = 0.001, Q_gyro = 0.005;
float R_angle = 0.5, C_0 = 1;
float dt = 10 * 0.001;
float angle_err, q_bias;
float Pdot[4] = {0, 0, 0, 0};
float P[2][2] = {{1, 0}, {0, 1}};
float PCt_0, PCt_1, E, K_0, K_1, t_0, t_1;
float gyro_m, Gyro_y, Gyro_z;
float accelz = 0;
int16_t ax, ay, az, gx, gy, gz;

#define MPU6050_RA_ACCEL_XOUT_H 0x3B
#define MPU6050_RA_GYRO_CONFIG 0x1B
#define MPU6050_RA_ACCEL_CONFIG 0x1C
#define MPU6050_RA_PWR_MGMT_1 0x6B
const char devAddr = 0x68;
/*
 * 变量,在中断函数中更新
 */

volatile float angle_a;  //当前小车倾斜角度
volatile float angle;
volatile float angle_dot;         //当前小车的加速度
volatile unsigned char mpuToken;  //令牌,代表已经进入过中断
unsigned char buffer[14];

ISR(TIMER1_COMPA_vect) {
    sei();

    Wire.beginTransmission(devAddr);
    Wire.write(MPU6050_RA_ACCEL_XOUT_H);
    Wire.endTransmission();
    Wire.beginTransmission(devAddr);
    if (Wire.requestFrom(devAddr, 14) != 14) return;
    for (int i = 0; i < 14; i++) buffer[i] = Wire.read();

    ax = (((int16_t)buffer[0]) << 8) | buffer[1];
    ay = (((int16_t)buffer[2]) << 8) | buffer[3];
    az = (((int16_t)buffer[4]) << 8) | buffer[5];
    gx = (((int16_t)buffer[8]) << 8) | buffer[9];
    gy = (((int16_t)buffer[10]) << 8) | buffer[11];
    gz = (((int16_t)buffer[12]) << 8) | buffer[13];
    //反着安装,az取反
    float angle_m = atan2(ay, -az) * 180 / PI;
    //反着安装,gx取反
    gyro_m = (-gx - 128.1) / 131;
    angle_a += (gyro_m - q_bias) * dt;
    angle_err = angle_m - angle_a;
    Pdot[0] = Q_angle - P[0][1] - P[1][0];
    Pdot[1] = -P[1][1];
    Pdot[2] = -P[1][1];
    Pdot[3] = Q_gyro;
    P[0][0] += Pdot[0] * dt;
    P[0][1] += Pdot[1] * dt;
    P[1][0] += Pdot[2] * dt;
    P[1][1] += Pdot[3] * dt;
    PCt_0 = C_0 * P[0][0];
    PCt_1 = C_0 * P[1][0];
    E = R_angle + C_0 * PCt_0;
    K_0 = PCt_0 / E;
    K_1 = PCt_1 / E;
    t_0 = PCt_0;
    t_1 = C_0 * P[0][1];
    P[0][0] -= K_0 * t_0;
    P[0][1] -= K_0 * t_1;
    P[1][0] -= K_1 * t_0;
    P[1][1] -= K_1 * t_1;
    angle_a += K_0 * angle_err;
    q_bias += K_1 * angle_err;
    angle_dot = gyro_m - q_bias;
    angle = angle_a - offset_angle;
    mpuToken++;
}
bool writeByte(uint8_t regAddr, uint8_t data) {
    Wire.beginTransmission(devAddr);
    Wire.write((uint8_t)regAddr);
    Wire.write((uint8_t)data);
    return Wire.endTransmission() == 0;
}
void mpuSetup() {
    Wire.begin();
    Wire.setClock(100000);
    while (!writeByte(MPU6050_RA_GYRO_CONFIG, 0))
        ;
    while (!writeByte(MPU6050_RA_ACCEL_CONFIG, 0))
        ;
    while (!writeByte(MPU6050_RA_PWR_MGMT_1, 1))
        ;
    // Timer1 CTC模式64分频 200hz = 5ms
    TCCR1A = 0;
    TCCR1B = 0;
    TCNT1 = 0;
    OCR1A = 16000000 / 64 / 100 - 1;
    TCCR1B = (1 << WGM12) | (1 << CS11) | (1 << CS10);
    TIMSK1 |= (1 << OCIE1A);
    sei();
}

/*
 * 反面安装时,姿态传感器需要一段时间才能找到正确的角度
 */