#include <EEPROM.h>

//以下参数适用于调试,此参数有着硬朗的直立参数和位置参数
//由于积分有有限幅,小车只能在不到轮子转一圈的范围内回复原点
float pidParameters[9] = {17, 1.0, 50, 600, 0.2, 0.4, 0.0, 1500};

//以下参数使用于遥控,通过减弱直立参数,减小遥控移动时的颤抖现象
//副作用:没有遥控信号时,摆动幅度会稍微增加
// float pidParameters[9] = {14.5, 1.0, 45, 600, 0.0, 0.4, 0.0, 1500};

//以下参数适用于爬坡,载物
//去掉了直立环和速度环的积分限幅,两者能之间相互抵消,
//从而让小车可以倾斜着保持平衡,载物时可以自适应重心
//使用此参数时,小车必须在平衡,并且轮子着地的情况下开机
// float pidParameters[8] = {18, 1.0, 55, 0, 0.2, 0.4, 0.0, 0};

#define pid_angle_kp (pidParameters[0])
#define pid_angle_ki (pidParameters[1])
#define pid_angle_kd (pidParameters[2])
#define pid_angle_I_limit (pidParameters[3])
#define pid_speed_kp (pidParameters[4])
#define pid_speed_ki (pidParameters[5])
#define pid_speed_kd (pidParameters[6])
#define pid_speed_I_limit (pidParameters[7])

extern volatile float angle;                   //从calcAngle文件引入angle变量
extern volatile long encodeCounterL;  //从encoder文件引入编码器计数变量
extern volatile long encodeCounterR;  //从encoder文件引入编码器计数变量
extern int rc_forward;                //从remote文件引入前进/后退期望
extern int rc_truen;                  //从remote文件引入转弯期望
extern float ultrPing;

int speedL = 0;            //左边要输出的pwm
int speedR = 0;            //右边要输出的pwm
long encl_latest = 0;      //记录上一次的编码器值,用以计算速度
long encr_latest = 0;      //记录上一次的编码器值,用以计算速度
float angle_integral = 0;  //角度积分
float angle_latest = 0;    //上一次的倾斜角度值
long spdL_integral = 0;    //速度积分
long spdL_latest = 0;      //上一次输出的速度
long spdR_integral = 0;    //速度积分
long spdR_latest = 0;      //上一次输出的速度

unsigned char output_counter = 0;

void resetPID() {
    angle_integral = 0;
    angle_latest = 0;
    spdL_integral = 0;
    spdL_latest = 0;
    spdR_integral = 0;
    spdR_latest = 0;
}
float rc_forward_filter = 0;
void output() {
    angle_integral += angle;
    //衰减角度积分,小车在小角度倾斜时,保持硬朗的直立(只衰减很小的数值)
    //在大角度倾斜时衰减较大的数值,让速度环来发挥作用
    angle_integral *= 0.99;

    if (pid_angle_I_limit != 0) {
        angle_integral =
            constrain(angle_integral, -pid_angle_I_limit, pid_angle_I_limit);
    }

    float angle_derr = angle - angle_latest;
    angle_latest = angle;
    //直立PID计算
    float pwmStand = pid_angle_kp * angle;
    pwmStand += pid_angle_ki * angle_integral;
    pwmStand += pid_angle_kd * angle_derr;

    //计算当前速度
    long enc_l = encodeCounterL;
    long enc_r = encodeCounterR;
    long spL = enc_l - encl_latest;
    long spR = enc_r - encr_latest;
    encl_latest = enc_l;
    encr_latest = enc_r;
    //叠加遥控信号
    //对遥控信号进行滤波,可以让小车平缓加速,减小对直立的影响
    rc_forward_filter = rc_forward_filter * 0.95 + 0.05 * rc_forward;
    spL -= rc_forward_filter * 0.15;
    spL -= rc_truen * 0.1;
    spR -= rc_forward_filter * 0.15;
    spR += rc_truen * 0.1;
    //速度积分
    spdL_integral += spL;
    spdR_integral += spR;
    if (pid_speed_I_limit != 0) {
        spdL_integral =
            constrain(spdL_integral, -pid_speed_I_limit, pid_speed_I_limit);
        spdR_integral =
            constrain(spdR_integral, -pid_speed_I_limit, pid_speed_I_limit);
    }
    //微分
    long splDerr = spL - spdL_latest;
    long sprDerr = spR - spdR_latest;
    spdL_latest = spL;
    spdR_latest = spR;
    //分别计算左右两边的pid
    float pwmSpeedL = pid_speed_kp * spL;
    pwmSpeedL += pid_speed_ki * spdL_integral;
    pwmSpeedL += pid_speed_kd * splDerr;

    float pwmSpeedR = pid_speed_kp * spR;
    pwmSpeedR += pid_speed_ki * spdR_integral;
    pwmSpeedR += pid_speed_kd * sprDerr;

    //速度融合,左轮叠加右边编码器,右轮叠加左边编码器
    //保持小车两个轮子位置同步
    speedL = pwmStand + pwmSpeedR;
    speedR = pwmStand + pwmSpeedL;
    //限制PWM
    speedL = constrain(speedL, -255, 255);
    speedR = constrain(speedR, -255, 255);

    //输出PWM
    if (abs(angle) < 40) {
        if (speedL < 0) {
            digitalWrite(12, LOW);  //左轮旋转方向为后退
        } else {
            digitalWrite(12, HIGH);  //左轮旋转方向为前进
        }
        int pwm = abs(speedL);
        if (pwm < 15) pwm = 0;  // pwm小于马达死区pwm,不输出pwm
        analogWrite(11, pwm);   //左轮pwm(速度)

        if (speedR < 0) {
            digitalWrite(8, HIGH);  //右轮旋转方向为后退
        } else {
            digitalWrite(8, LOW);  //右轮旋转方向为前进
        }
        pwm = abs(speedR);
        if (pwm < 15) pwm = 0;
        analogWrite(3, pwm);  //右轮pwm(速度)
    } else {
        //如果倾斜角度太大,停止输出
        analogWrite(3, 0);
        analogWrite(11, 0);
        resetPID();
    }
    // output函数执行周期为100hz(10毫秒),时间非常紧
    // 执行打印程序会影响整体稳定性
    if (output_counter % 16 == 0) {
        Serial.print(angle);  //打印当前倾斜角度
        // Serial.print(',');
        // Serial.print(spL);  //打印当前左轮速度
        // Serial.print(',');
        // Serial.print(spR);  //打印当前右轮速度
        // Serial.print(',');
        // Serial.print(speedL);  //打印当前左轮pwm
        // Serial.print(',');
        // Serial.print(speedR);  //打印当前右轮pwm
        // Serial.print(',');
        // Serial.print(encodeCounterL);  //打印左轮编码器值
        // Serial.print(',');
        // Serial.print(encodeCounterR);  //打印右轮编码器值
        // Serial.print(',');
        // Serial.print(angle_integral);  //打印当前角度积分
        // Serial.print(',');
        // Serial.print(spdL_integral); //打印当前左轮速度积分(速度的积分是位置)
        // Serial.print(',');
        // Serial.print(spdR_integral); //打印当前右轮速度积分(速度的积分是位置)
        // Serial.print(',');
        // Serial.print(ultrPing);  //打印当前超声波测量读数

        //不打印数据也打印换行,这样可以通过观察TX灯判断当前程序是否在执行
        Serial.println();
    }

    output_counter++;
}

void pidSetup() {
    pinMode(12, OUTPUT);
    pinMode(11, OUTPUT);
    pinMode(8, OUTPUT);
    pinMode(3, OUTPUT);
    if (EEPROM.read(0) != 'P') return;
    if (EEPROM.read(1) != 'I') return;
    if (EEPROM.read(2) != 'D') return;
    for (int p = 0; p < sizeof(pidParameters); p++) {
        ((uint8_t*)pidParameters)[p] = EEPROM.read(3 + p);
    }
    for (int i = 0; i < sizeof(pidParameters) / sizeof(pidParameters[0]); i++) {
        Serial.print(pidParameters[i]);
        Serial.print(',');
    }
    Serial.println();
}
//串口监视器输入  p18, 1.0, 55, 600, 0.2, 0.4, 0.0, 1500 [换行/回车]
//可以运行时修改pid参数,并保持到eeprom中
//设置之后,程序中定义的参数将不再起作用
void parseParameters() {
    uint8_t i = 0;
    analogWrite(3, 0);
    analogWrite(11, 0);
    while (1) {
        float a = Serial.parseFloat();
        wdt_reset();
        if (i < sizeof(pidParameters) / sizeof(pidParameters[0])) {
            pidParameters[i] = a;
        }
        i++;
        Serial.println(a);
        if (Serial.read() == ',') continue;
        EEPROM.write(0, 'P');
        EEPROM.write(1, 'I');
        EEPROM.write(2, 'D');
        for (int p = 0; p < sizeof(pidParameters); p++) {
            wdt_reset();
            EEPROM.write(3 + p, ((uint8_t*)pidParameters)[p]);
        }
        resetPID();
        break;
    }
}